@@ -13,13 +13,14 @@ use std::ffi::CString;
 use std::fmt;
 use std::os::raw::{c_char, c_int};
 use std::os::unix::io::AsRawFd;
+use std::pin::Pin;
 use std::ptr;
-use std::rc::Rc;
+use std::sync::{Arc, Mutex};
 
 /// Occlum-PAL error codes.
 #[derive(Debug, Clone, Copy, Eq, PartialEq)]
 pub enum Error {
-    PalApiError,
+    VersionError,
     InitError,
     CreateError,
     ExecError,
@@ -30,7 +31,7 @@ pub enum Error {
 impl fmt::Display for Error {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         match self {
-            Error::PalApiError => write!(f, "PAL API version mismatch"),
+            Error::VersionError => write!(f, "PAL API version mismatch"),
             Error::InitError => write!(f, "Initialization error"),
             Error::CreateError => write!(f, "Process creation error"),
             Error::ExecError => write!(f, "Process execution error"),
@@ -55,11 +56,11 @@ pub enum LogLevel {
 impl LogLevel {
     fn as_bytes(&self) -> &'static [u8] {
         match *self {
-            LogLevel::Off => b"off\0",
-            LogLevel::Error => b"error\0",
-            LogLevel::Warn => b"warn\0",
-            LogLevel::Info => b"info\0",
-            LogLevel::Trace => b"trace\0",
+            LogLevel::Off => b"off",
+            LogLevel::Error => b"error",
+            LogLevel::Warn => b"warn",
+            LogLevel::Info => b"info",
+            LogLevel::Trace => b"trace",
         }
     }
 }
@@ -94,15 +95,6 @@ impl Default for Config {
     }
 }
 
-impl Config {
-    fn to_api(&self) -> occlum_pal_attr {
-        occlum_pal_attr {
-            instance_dir: self.instance_dir.as_ptr() as *const c_char,
-            log_level: self.log_level.as_bytes().as_ptr() as *const c_char,
-        }
-    }
-}
-
 #[derive(Debug, Clone, PartialEq, Eq)]
 pub struct Stdio {
     stdin: c_int,
@@ -130,12 +122,12 @@ impl Stdio {
         }
     }
 
-    fn to_api(&self) -> occlum_stdio_fds {
-        occlum_stdio_fds {
+    fn to_api(&self) -> Pin<Box<occlum_stdio_fds>> {
+        Box::pin(occlum_stdio_fds {
             stdin_fd: self.stdin,
             stdout_fd: self.stdout,
             stderr_fd: self.stderr,
-        }
+        })
     }
 }
 
@@ -144,35 +136,70 @@ impl Stdio {
 pub struct ProcessBuilder {
     path: CString,
     argv: Vec<CString>,
-    argv_api: Option<Vec<*const c_char>>,
     env: Vec<CString>,
-    env_api: Option<Vec<*const c_char>>,
     pid: c_int,
     stdio: Option<Stdio>,
 }
 
+#[derive(Debug)]
+pub struct CStringsVec {
+    cstrings: Vec<CString>,
+    ptrs: Vec<*const c_char>,
+}
+
+impl CStringsVec {
+    fn new(cstrings: Vec<CString>) -> Result<Self, Error> {
+        let mut ptrs = Vec::with_capacity(cstrings.len() + 1);
+        for cstring in &cstrings {
+            ptrs.push(cstring.as_ptr());
+        }
+        ptrs.push(ptr::null());
+        Ok(Self { cstrings, ptrs })
+    }
+
+    fn as_mut_ptr(&mut self) -> *mut *const c_char {
+        self.ptrs.as_mut_ptr()
+    }
+}
+
+#[derive(Debug)]
+struct ProcessApiConcrete {
+    path: CString,
+    argv_cstrings_vec: CStringsVec,
+    env_cstrings_vec: CStringsVec,
+    stdio_api: Pin<Box<occlum_stdio_fds>>,
+    pid: Pin<Box<c_int>>,
+}
+
+#[derive(Debug)]
+struct ProcessApi {
+    concrete: ProcessApiConcrete,
+    ptrs: Pin<Box<occlum_pal_create_process_args>>,
+}
+
 /// A process to be run in an enclave.
 #[derive(Debug)]
 pub struct Process {
-    builder: ProcessBuilder,
-    enclave: Rc<Enclave>,
-    pid: c_int,
+    enclave: Arc<Mutex<Enclave>>,
+    process_api: ProcessApi,
 }
 
 impl Process {
     /// Return the process identifier.
     pub fn process_id(&self) -> ProcessId {
-        ProcessId { pid: self.pid }
+        ProcessId {
+            pid: unsafe { *self.process_api.ptrs.pid },
+        }
     }
 
     /// Execute the process inside the Occlum enclave
     pub fn exec(&self) -> Result<ExitCode, Error> {
         let mut exit_code: c_int = -1;
-        let mut exec_args = occlum_pal_exec_args {
-            pid: self.pid,
+        let mut exec_args = Box::pin(occlum_pal_exec_args {
+            pid: self.process_id().pid,
             exit_value: &mut exit_code,
-        };
-        let exec_result = unsafe { occlum_pal_exec(&mut exec_args) };
+        });
+        let exec_result = unsafe { occlum_pal_exec(&mut *exec_args) };
         if exec_result == 0 {
             Ok(exit_code)
         } else {
@@ -250,69 +277,91 @@ impl ProcessBuilder {
         Ok(Self {
             path,
             argv,
-            argv_api: None,
             env,
-            env_api: None,
             pid: -1,
             stdio,
         })
     }
 
     /// Create a new process to be run in the enclave.
-    pub fn build(mut self, enclave: &Rc<Enclave>) -> Result<Process, Error> {
-        let mut argv_api: Vec<*const c_char> = vec![];
-        for x in &self.argv {
-            argv_api.push(x.as_ptr());
-        }
-        argv_api.push(ptr::null());
-
-        let mut env_api: Vec<*const c_char> = vec![];
-        for x in &self.env {
-            env_api.push(x.as_ptr());
-        }
-        env_api.push(ptr::null());
-
-        self.argv_api = Some(argv_api);
-        self.env_api = Some(env_api);
-
-        let stdio = Stdio::default().to_api();
-
-        let mut process_api = occlum_pal_create_process_args {
-            path: self.path.as_ptr(),
-            argv: self.argv_api.as_mut().unwrap().as_mut_ptr(),
-            env: self.env_api.as_mut().unwrap().as_mut_ptr(),
-            stdio: &stdio,
-            pid: &mut self.pid,
+    pub fn build(self, enclave: &Arc<Mutex<Enclave>>) -> Result<Process, Error> {
+        let stdio_api = match self.stdio {
+            None => Stdio::default().to_api(),
+            Some(stdio) => stdio.to_api(),
         };
-
-        if unsafe { occlum_pal_create_process(&mut process_api) } != 0 {
+        let mut process_api_concrete = ProcessApiConcrete {
+            path: self.path,
+            argv_cstrings_vec: CStringsVec::new(self.argv)?,
+            env_cstrings_vec: CStringsVec::new(self.env)?,
+            stdio_api,
+            pid: Box::pin(-1),
+        };
+        let mut process_api_ptrs = Box::pin(occlum_pal_create_process_args {
+            path: process_api_concrete.path.as_ptr(),
+            argv: process_api_concrete.argv_cstrings_vec.as_mut_ptr(),
+            env: process_api_concrete.env_cstrings_vec.as_mut_ptr(),
+            stdio: &*process_api_concrete.stdio_api,
+            pid: &mut *process_api_concrete.pid,
+        });
+        if unsafe { occlum_pal_create_process(&mut *process_api_ptrs) } != 0 {
             return Err(Error::CreateError);
         }
-        let pid = unsafe { *process_api.pid };
         Ok(Process {
-            builder: self,
             enclave: enclave.clone(),
-            pid,
+            process_api: ProcessApi {
+                concrete: process_api_concrete,
+                ptrs: process_api_ptrs,
+            },
         })
     }
 }
 
+#[derive(Debug)]
+struct ConfigApiConcrete {
+    instance_dir: CString,
+    log_level: CString,
+}
+
+#[derive(Debug)]
+struct ConfigApi {
+    concrete: ConfigApiConcrete,
+    ptrs: Pin<Box<occlum_pal_attr>>,
+}
+
 /// An Occlum SGX enclave.
 #[derive(Debug)]
 pub struct Enclave {
     config: Config,
+    config_api: ConfigApi,
 }
 
 impl Enclave {
     /// Create a new SGX enclave with the given configuration.
-    pub fn new(config: Config) -> Result<Rc<Self>, Error> {
-        if (unsafe { occlum_pal_get_version() } <= 0) {}
+    pub fn new(config: Config) -> Result<Arc<Mutex<Self>>, Error> {
+        if (unsafe { occlum_pal_get_version() } <= 0) {
+            return Err(Error::VersionError);
+        }
 
-        let config_api = config.to_api();
-        if unsafe { occlum_pal_init(&config_api) } != 0 {
+        let config_api_concrete = ConfigApiConcrete {
+            instance_dir: CString::new(config.instance_dir.as_bytes())
+                .map_err(|_| Error::CStringError)?,
+            log_level: CString::new(config.log_level.as_bytes())
+                .map_err(|_| Error::CStringError)?,
+        };
+        let config_api_ptrs = Box::pin(occlum_pal_attr {
+            instance_dir: config_api_concrete.instance_dir.as_ptr() as *const c_char,
+            log_level: config_api_concrete.log_level.as_bytes().as_ptr() as *const c_char,
+        });
+        if unsafe { occlum_pal_init(&*config_api_ptrs) } != 0 {
             return Err(Error::InitError);
         }
-        Ok(Rc::new(Enclave { config }))
+        Ok(Arc::new(Mutex::new(Enclave {
+            config,
+            config_api: ConfigApi {
+                concrete: config_api_concrete,
+                ptrs: config_api_ptrs,
+            },
+        })))
     }
 
     fn destroy(&mut self) {
